/* 
   Main game functions
   --------------------------------------------------------------------
   Linberto

   Copyright (C) 1999-2001 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "game.h"

static int evil_sound; /* Which evil sound I must play? */

void game_loop( void)
{
 char buffer[ 256];

 stage     = opt.cfg_start_stage;
 level     = 0;
 lives     = opt.cfg_lives;
 score     = 0;

 sound_music( SND_BKGDM);

 game_init_actors();     /* Initialize actor structures */
 
 game_init_map_stage( opt.cfg_world_path);

 game_show_stage();      /* Show a welcome message      */

 /* Draw the game cubes */
 draw_berto_board( (int) map_stage[ act_linberto.mapy][ act_linberto.mapx]);

 /* Copy the graphics to a buffer screen */
 gl_copyscreen( backscreen);

 /* Put the Linberto sprite & end the screen mounting */
 game_move_beginning();
 game_show_text();

 game_set_evil_timers( &act_bubble   );
 game_set_evil_timers( &act_flowerpot);

 while( 1) {

       if( lives == 0) {
           /* Game over */
           game_over();
           break;
       }

       if( game_check_complete() == 1) {
           /* Show complete level message */
           sprintf( buffer, _(" Level %d completed! "), level);
           screen_center_message( buffer, LIN_RED);
           score += 100;
           level++;

           if( stage == max_stages - 1 && level == MAX_LEVEL) {
               game_win();
               return;
           }

           if( level == MAX_LEVEL) {
               stage++;
               level = 0;
               lives++;
               score += 1000;
               screen_center_message( _(" STAGE CLEAR "), LIN_RED);
           }

           game_init_actors();
           game_init_map_stage( opt.cfg_world_path);

           if( level == 0)
               game_show_stage();

           gl_clearscreen( 0);
           draw_berto_board( (int) map_stage[ act_linberto.mapy][ act_linberto.mapx]);
           gl_copyscreen( backscreen);
           game_move_beginning();
           game_show_text();
           game_set_evil_timers( &act_bubble   );
           game_set_evil_timers( &act_flowerpot);
           continue;
       }

       evil_start_x = BUB_START_X;
       evil_start_y = BUB_START_Y;
       evil_sound   = SND_BUBJP;
       if( game_check_evils( &act_bubble   ) < 0)
           continue;

       evil_start_x = FLP_START_X;
       evil_start_y = FLP_START_Y;
       evil_sound   = SND_FPJP;
       if( game_check_evils( &act_flowerpot) < 0)
           continue;

       if( keyboard_update() == 0 && joy_moved() == 0)
           continue;

       switch( lin_scancode) {
               case SCANCODE_Q:
               case SCANCODE_ESCAPE:
                      score_high_check();
                      return;
               case SCANCODE_P:
               case SCANCODE_SPACE:
                      game_pause();
                      continue;
               case SCANCODE_CURSORLEFT:
               case SCANCODE_CURSORBLOCKLEFT:
               case SCANCODE_KEYPAD1:
                      act_linberto.mapx--;
                      act_linberto.mapy++;
                      if( game_checkpos() < 0)
                          continue;
                      move_sprite_downleft();
                      game_draw_block();
                      continue;
               case SCANCODE_CURSORRIGHT:
               case SCANCODE_CURSORBLOCKRIGHT:
               case SCANCODE_KEYPAD9:
                      act_linberto.mapx++;
                      act_linberto.mapy--;
                      if( game_checkpos() < 0)
                          continue;
                      move_sprite_upright();
                      game_draw_block();
                      continue;
               case SCANCODE_CURSORUP:
               case SCANCODE_CURSORBLOCKUP:
               case SCANCODE_KEYPAD7:
                      act_linberto.mapx--;
                      act_linberto.mapy--;
                      if( game_checkpos() < 0)
                          continue;
                      move_sprite_upleft();
                      game_draw_block();
                      continue;
               case SCANCODE_CURSORDOWN:
               case SCANCODE_CURSORBLOCKDOWN:
               case SCANCODE_KEYPAD3:
                      act_linberto.mapx++;
                      act_linberto.mapy++;
                      if( game_checkpos() < 0)
                          continue;
                      move_sprite_downright();
                      game_draw_block();
                      continue;
               case SCANCODE_S:
                      game_sound();
                      continue;
               case SCANCODE_I:
                      opt.cfg_showtext = opt.cfg_showtext == SHOWTEXT_ON ? SHOWTEXT_OFF : SHOWTEXT_ON;
                      game_show_text();
                      continue;
       } /* end switch */

 } /* end while( 1) */

 score_high_check();
}

void game_init_actors( void)
{
 game_init_linberto(  ACTOR_ACTIVE  );
 game_init_bubble(    ACTOR_INACTIVE);
 game_init_flowerpot( ACTOR_INACTIVE);
}

void game_init_linberto( int active_status)
{
 act_linberto.mapx   = 7;
 act_linberto.mapy   = 0;
 act_linberto.active = active_status;
 act_linberto.sp     = &sprite_linberto; 
 act_linberto.sp->x  = LIN_START_X;
 act_linberto.sp->y  = LIN_START_Y;
 act_linberto.lp.lx  = LIN_START_X;
 act_linberto.lp.ly  = LIN_START_Y;
 act_linberto.lp.lw  = sprite_linberto.width;
 act_linberto.lp.lh  = sprite_linberto.height;
}

void game_init_bubble( int active_status)
{
 act_bubble.mapx   = 7;
 act_bubble.mapy   = 0;
 act_bubble.active = active_status;
 act_bubble.sp     = &sprite_bubble;
 act_bubble.sp->x  = BUB_START_X;
 act_bubble.sp->y  = BUB_START_Y;
 act_bubble.lp.lx  = BUB_START_X;
 act_bubble.lp.ly  = BUB_START_Y;
 act_bubble.lp.lw  = sprite_bubble.width;
 act_bubble.lp.lh  = sprite_bubble.height;
 act_bubble.wait_clockticks[ 0] = 100 * opt.cfg_speed;
 act_bubble.wait_clockticks[ 1] =  90 * opt.cfg_speed;
 act_bubble.wait_clockticks[ 2] =  80 * opt.cfg_speed;
 act_bubble.wait_clockticks[ 3] =  60 * opt.cfg_speed;
 act_bubble.wait_clockticks[ 4] =  30 * opt.cfg_speed;
 act_bubble.init_func = &game_init_bubble;
 act_bubble.kill_func = &game_linberto_bubble_trapped;
}

void game_init_flowerpot( int active_status)
{
 act_flowerpot.mapx   = 7;
 act_flowerpot.mapy   = 0;
 act_flowerpot.active = active_status;
 act_flowerpot.sp     = &sprite_flowerpot;
 act_flowerpot.sp->x  = FLP_START_X;
 act_flowerpot.sp->y  = FLP_START_Y;
 act_flowerpot.lp.lx  = FLP_START_X;
 act_flowerpot.lp.ly  = FLP_START_Y;
 act_flowerpot.lp.lw  = sprite_flowerpot.width;
 act_flowerpot.lp.lh  = sprite_flowerpot.height;
 act_flowerpot.wait_clockticks[ 0] = 200 * opt.cfg_speed;
 act_flowerpot.wait_clockticks[ 1] = 180 * opt.cfg_speed;
 act_flowerpot.wait_clockticks[ 2] = 160 * opt.cfg_speed;
 act_flowerpot.wait_clockticks[ 3] = 120 * opt.cfg_speed;
 act_flowerpot.wait_clockticks[ 4] =  60 * opt.cfg_speed;
 act_flowerpot.init_func = &game_init_flowerpot;
 act_flowerpot.kill_func = &game_linberto_flowerpot_kill;
}

void game_init_map_stage( char *filename)
{
 FILE *fp;
 int i, j;
 char worldname[ 2048];

 sprintf( worldname, "%s/%s", LINBERTO_DATADIR, filename);

 fp = linberto_openfile( worldname, LINBERTO_WORLDFILE_DEFAULT, "r");

 fseek( fp, ( long) sizeof( struct world_header    ) + stage * 
                    sizeof( map_stage), SEEK_SET);

 fread( &map_stage, sizeof( map_stage), 1, fp);

 fclose( fp);

 for( i = 0; i < MAP_X_NDX; i++)
      for( j = 0; j < MAP_Y_NDX; j++)
           if( map_stage[ i][ j] != -1)
               map_stage[ i][ j] = level + 1;
           else
               map_stage[ i][ j] = -1;

}

void game_move_beginning( void)
{
 act_linberto.sp->x = LIN_START_X;
 act_linberto.sp->y = LIN_START_Y;
 move_sprite();
}

void move_sprite( void)
{
 int x1, y1, w1, h1;
 int x2, y2, w2, h2;
 int x3, y3, w3, h3;

 /* Set the sprite screen as the current context */
 gl_setcontext( spritescreen);

 if( act_linberto.active  == ACTOR_ACTIVE) {
     misc_calculate_minmax( &act_linberto , &x1, &y1, &w1, &h1);
     /* Copy the cubes from a sprite clean 'backscreen' context */
     gl_copyboxfromcontext( backscreen    , x1, y1, w1, h1, x1, y1);
 }

 if( act_bubble.active    == ACTOR_ACTIVE) {
     misc_calculate_minmax( &act_bubble   , &x2, &y2, &w2, &h2);
     gl_copyboxfromcontext( backscreen    , x2, y2, w2, h2, x2, y2);
 }

 if( act_flowerpot.active == ACTOR_ACTIVE) {
     misc_calculate_minmax( &act_flowerpot, &x3, &y3, &w3, &h3);
     gl_copyboxfromcontext( backscreen    , x3, y3, w3, h3, x3, y3);
 }

 /* Put the sprites on screen */
 if( act_linberto.active  == ACTOR_ACTIVE)
     sprite_put_mask( act_linberto.sp );
 if( act_bubble.active    == ACTOR_ACTIVE)
     sprite_put_mask( act_bubble.sp   );
 if( act_flowerpot.active == ACTOR_ACTIVE)
     sprite_put_mask( act_flowerpot.sp);

 /* Now copy the image to the physical screen */
 gl_setcontext( physicalscreen);

 if( act_linberto.active  == ACTOR_ACTIVE) {
     gl_copyboxfromcontext( spritescreen, x1, y1, w1, h1, x1, y1);
     misc_save_last_position( &act_linberto);
 }

 if( act_bubble.active    == ACTOR_ACTIVE) {
     gl_copyboxfromcontext( spritescreen, x2, y2, w2, h2, x2, y2);
     misc_save_last_position( &act_bubble);
 }

 if( act_flowerpot.active == ACTOR_ACTIVE) {
     gl_copyboxfromcontext( spritescreen, x3, y3, w3, h3, x3, y3);
     misc_save_last_position( &act_flowerpot);
 }

}

void move_sprite_downleft( void)
{
 sound_send( SND_JUMP);
 act_linberto.sp->x -= 16;
 act_linberto.sp->y += 21;
 move_sprite();
}

void move_sprite_downright( void)
{
 sound_send( SND_JUMP);
 act_linberto.sp->x += 16;
 act_linberto.sp->y += 21;
 move_sprite();
}

void move_sprite_upleft( void)
{
 sound_send( SND_JUMP);
 act_linberto.sp->x -= 16;
 act_linberto.sp->y -= 21;
 move_sprite();
}

void move_sprite_upright( void)
{
 sound_send( SND_JUMP);
 act_linberto.sp->x += 16;
 act_linberto.sp->y -= 21;
 move_sprite();
}

void game_linberto_died( void)
{
 /* Show death message if we are not in INVINCIBLE MODE */
 if( opt.cfg_invincible == INVINCIBLE_OFF) {
     sound_send( SND_DEATH);
     lives--;
     if( lives >= 1)
         screen_center_message( _(" Life lost "), LIN_RED);
 }

 game_delete_sprite( &act_linberto );
 game_delete_sprite( &act_bubble   );
 game_delete_sprite( &act_flowerpot);

 /* This is needed to delete the flowerpot kill
    if Linberto died because the flowerpot      */
 act_flowerpot.sp->x = act_flowerpot.sp->x + FLP_FPK_X;
 act_flowerpot.sp->y = act_flowerpot.sp->y + FLP_FPK_Y;
 game_delete_sprite( &act_flowerpot);

 game_init_actors();
 game_move_beginning();
 game_show_text();

 game_set_evil_timers( &act_bubble   );
 game_set_evil_timers( &act_flowerpot);
} 

int game_checkpos( void)
{

 if( act_linberto.mapy < 0 || act_linberto.mapy > MAP_X_NDX - 1) {
     sound_send( SND_FALL);
     if( opt.cfg_sound == SOUND_ON)
         misc_wait( 100, WAIT_ESC_ENTER);
     game_linberto_died();
     return -1;
 }

 if( act_linberto.mapx < 0 || act_linberto.mapx > MAP_Y_NDX - 1) {
     sound_send( SND_FALL);
     if( opt.cfg_sound == SOUND_ON)
         misc_wait( 100, WAIT_ESC_ENTER);
     game_linberto_died();
     return -1;
 }

 if( map_stage[ act_linberto.mapy][ act_linberto.mapx] == -1) {
     sound_send( SND_FALL);
     if( opt.cfg_sound == SOUND_ON)
         misc_wait( 100, WAIT_ESC_ENTER);
     game_linberto_died();
     return -1;
 }

 if( act_bubble.active == ACTOR_ACTIVE)
     if( act_linberto.mapx == act_bubble.mapx && act_linberto.mapy == act_bubble.mapy) {
         game_linberto_bubble_trapped();
         game_linberto_died();
         return -1;
     }

 if( act_flowerpot.active == ACTOR_ACTIVE)
     if( act_linberto.mapx == act_flowerpot.mapx && act_linberto.mapy == act_flowerpot.mapy) {
         game_linberto_flowerpot_kill();
         game_linberto_died();
         return -1;
     }

 if( map_stage[ act_linberto.mapy][ act_linberto.mapx] != 0) {
     map_stage[ act_linberto.mapy][ act_linberto.mapx]--;
     score += 10;
     game_show_text();
 }

 return 0;
}

int game_check_complete( void)
{
 int i, j;
 char c;

 for( i = 0; i < MAP_X_NDX; i++)
      for( j = 0; j < MAP_Y_NDX; j++) {
           c = map_stage[ i][ j];
           if( c > 0) /* c != -1 &&  c != 0 */
               return 0;  
      }
 
 return 1;
}

void game_draw_block( void)
{
 int x, y;

 x = act_linberto.sp->x +  9;
 y = act_linberto.sp->y + 17;

 gl_setcontext( backscreen);
 draw_block( x, y, lcolors[ ( int) map_stage[ act_linberto.mapy][ act_linberto.mapx]]);

 gl_setcontext( physicalscreen);
 gl_copyboxfromcontext( backscreen, x - 16, y, 32, 29, x - 16, y);

 move_sprite();
}

void game_show_text( void)
{
 if( opt.cfg_showtext) {
     grf_printf(   0, 190, _("Level: %-1d"), level );
     grf_printf(  90, 190, _("Lives: %-2d"), lives );
     grf_printf( 180, 190, _("Score: %-8ld"), score );
     if( opt.cfg_sound == SOUND_ON)
         gl_write( 0, 0, "\16");
     else {
         gl_write( 0, 0, "\16");
         gl_line( 0, 0, 7, 7, LIN_RED);
         gl_line( 7, 0, 0, 7, LIN_RED);
     }
 } else {
     gl_fillbox( 0,   0,   8,  8, LIN_BLACK);
     gl_fillbox( 0, 190, 320, 10, LIN_BLACK);
 }

}

int game_check_evils( struct actor_struct *act)
{
 struct tms buf;
 clock_t waitime;

 waitime = times( &buf);

 if( opt.cfg_invincible == INVINCIBLE_ON)
     return 1;

 if( ( waitime - act->start_time) > act->wait_clockticks[ level]) {
 
     if( act->active == ACTOR_INACTIVE) {
         act->init_func( ACTOR_ACTIVE);

         move_sprite();

         goto check_pos;
     }

     switch( game_where_to_move( act)) {
             case MOVE_START: act->sp->x = evil_start_x;
                              act->sp->y = evil_start_y;
                              move_sprite();
                              break;
             case MOVE_DR   : act->sp->x += 16;
                              act->sp->y += 21;
                              sound_send( evil_sound);
                              move_sprite();
                              break;
             case MOVE_DL   : act->sp->x -= 16;
                              act->sp->y += 21;
                              sound_send( evil_sound);
                              move_sprite();
                              break;
             case MOVE_UR   : act->sp->x += 16;
                              act->sp->y -= 21;
                              sound_send( evil_sound);
                              move_sprite();
                              break;
             case MOVE_UL   : act->sp->x -= 16;
                              act->sp->y -= 21;
                              sound_send( evil_sound);
                              move_sprite();
                              break;
     }

check_pos:
     /* If linberto and the evil are in the same
        place, then linberto will die */
     if( act_linberto.mapx == act->mapx && act_linberto.mapy == act->mapy) {
         act->kill_func();
         game_linberto_died();
         return -1;
     }

     game_set_evil_timers( act);
     return -1;
 }

 return 1;
}

int game_where_to_move( struct actor_struct *act)
{
 struct timeval random_seed;
 int zero_or_one;
 int x, y;
 x = act->mapx;
 y = act->mapy;

 if( act_linberto.mapx == act->mapx && act_linberto.mapy == act->mapy)
     return MOVE_NONE;

 if( opt.cfg_enemykind == ENEMY_DUMB) {
     gettimeofday( &random_seed, ( struct timezone *) NULL);
     srand( (unsigned int) random_seed.tv_usec);
     zero_or_one = rand() % 2;

     act->mapy++;
     if( act->mapy > 6) {
         act->mapx = 7;
         act->mapy = 0;
         return MOVE_START;
     }

     if( zero_or_one) {
         act->mapx++;
         return MOVE_DR;
     } else {
         act->mapx--;
         return MOVE_DL;
     }

 }

 if( act_linberto.mapx <= act->mapx && act_linberto.mapy <= act->mapy) {
     act->mapx--;
     act->mapy--;
     if( act->mapx < 0 || act->mapy < 0) {
         act->mapx = x;
         act->mapy = y;
         act->mapx--;
         act->mapy++;
         return MOVE_DL;
     }
     return MOVE_UL;
 }

 if( act_linberto.mapx >= act->mapx && act_linberto.mapy > act->mapy) {
     act->mapx++;
     act->mapy++;
     return MOVE_DR;
 }

 if( act_linberto.mapx < act->mapx && act_linberto.mapy > act->mapy) {
     act->mapx--;
     act->mapy++;
     return MOVE_DL;
 }

 if( act_linberto.mapx > act->mapx && act_linberto.mapy <= act->mapy) {
     act->mapx++;
     act->mapy--;
     if( act->mapy < 0) {
         act->mapy = y;
         act->mapy++;
         return MOVE_DR;
     }
     return MOVE_UR;
 }

 return MOVE_NONE;
}

void game_linberto_bubble_trapped( void)
{
 game_delete_sprite( &act_linberto );
 game_delete_sprite( &act_bubble   );
 game_delete_sprite( &act_flowerpot);

 sound_send( SND_BUBK);

 sprite_bubble_trap.x = sprite_bubble.x;
 sprite_bubble_trap.y = sprite_bubble.y;
 sprite_put_mask( &sprite_bubble_trap);
 misc_wait( 100, WAIT_ESC_ENTER);
}

void game_linberto_flowerpot_kill( void)
{
 game_delete_sprite( &act_linberto );
 game_delete_sprite( &act_bubble   );
 game_delete_sprite( &act_flowerpot);

 sound_send( SND_SPLAT);

 sprite_pot_kill.x = sprite_flowerpot.x + FLP_FPK_X;
 sprite_pot_kill.y = sprite_flowerpot.y + FLP_FPK_Y;
 sprite_put_mask( &sprite_flowerpot);
 sprite_put_mask( &sprite_pot_kill);
 misc_wait( 100, WAIT_ESC_ENTER);
}

/*
 * Delete the sprite from the cube board
 */

void game_delete_sprite( struct actor_struct *act)
{
 int x, y, w, h;

 gl_setcontext( spritescreen);
 
 misc_calculate_minmax( act , &x, &y, &w, &h);

 gl_copyboxfromcontext( backscreen, x, y, w, h, x, y);

 gl_setcontext( physicalscreen);

 gl_copyboxfromcontext( spritescreen, x, y, w, h, x, y);
}

/*
 * Show a starting screen for each stage
 */
void game_show_stage( void)
{
 char buffer[ 256];
 int x, len;

 gl_clearscreen( 0);

 /* Draw a mini version of the stage map */
 draw_mini_board( 18, 10);

 /* Print the stage text information centered at the right side */
 len = 8 * strlen( _(" Stage %d ")) + 8;
 x = (( 190 - len) / 2) + 130;
 grf_printf( x, 38, _(" Stage %d "), stage);

 sprintf( buffer, _(" Level %d    Lives %d "), level, lives);
 screen_center_message( buffer, 0);

 gl_clearscreen( 0);
}

void game_set_evil_timers( struct actor_struct *act)
{
 struct tms buf;
 act->start_time = times( &buf);
}

void game_win( void)
{
 /* Show end game */
 screen_center_message( _(" YOU WIN!!! "), LIN_RED);
 /* Check score */
 score_high_check();
}

void game_pause( void)
{
 struct tms buf;
 clock_t bub_currtime; /* Bubble    */
 clock_t bub_stoptime; 
 clock_t flo_currtime; /* Flowerpot */
 clock_t flo_stoptime;

 /*
  * We need to stop the evil timers during the
  * pause or they will kill us when the pause is
  * over.
  */

 bub_currtime = times( &buf);
 flo_currtime = bub_currtime;

 bub_stoptime = bub_currtime - act_bubble.start_time;
 flo_stoptime = flo_currtime - act_flowerpot.start_time;

 screen_center_message( _(" Game paused "), LIN_RED);

 bub_currtime = times( &buf);
 flo_currtime = bub_currtime;

 act_bubble.start_time    = bub_currtime - bub_stoptime;
 act_flowerpot.start_time = flo_currtime - flo_stoptime;
}

void game_sound( void)
{
 opt.cfg_sound = opt.cfg_sound == SOUND_ON ? SOUND_OFF : SOUND_ON;

 if( opt.cfg_sound == SOUND_OFF) {
     opt.cfg_sound = SOUND_ON;
     sound_music( SND_NONE);
     opt.cfg_sound = SOUND_OFF;
 }
 else
     sound_music( SND_BKGDM);

 game_show_text();
}

void game_over( void)
{
 /* Show game over message */
 screen_center_message( _(" Game Over "), LIN_RED);
}

